unit pasmysql;

{$IFDEF FPC}
{$MODE Delphi}
{$H+}
{ELSE}
  {$IFNDEF LINUX}
  {$DEFINE WIN32}
  {$ENDIF}
{$ENDIF}

//work toward version 4.0:
{$DEFINE MYSQL4}

//I try to maintain backward (=3.23) compatability...

interface
uses
//{$IFNDEF LINUX}{$IFDEF FPC}LCLIntf,{$ENDIF}{$ENDIF}
     Classes, SysUtils,
     libmysql,
     passql,
     sqlsupport;

//Now doing a big revision
//Done
// * Add support for MySQL 4.x
// Merged TLiteDB and TMyDB

// ToDo:
// * More support functions like:
// QueryInteger QueryFloat QueryOneInteger etc.
// QueryTStrings, QueryCombobox (options[addempty, keeptext etc])
// etc etc
// Check licensensing conditions, i'll might not use LGPL but MPL or BSDPL for example




  //Some mathematic constants to support float functions
  //with delphi < 6 . delphi6 & kylix got these constants in the math unit.


//v0.2 Curently developing on w2kpro delphi5
//Added loads of stuff, will release v. zero . 1 1/2

//v0.1
//i'm developing on Linux Redhat 7.1 with
//custom installation of MySQL 3.23 client+server,
//typically i need the new libs.
//Obviously there's no way to check library version before compiling :(


////////////////////////////////////////////////
//                                            //
// TMyDB component v 0.1 by artee@dubaron.com //
// TMyDB is a MySQL specific interface        //
//                                            //
////////////////////////////////////////////////

// MySQL stand-alone interface unit & component
// Developed with Kylix OE,
// should work on delphi 4 and higher as well
// By artee@dubaron.com
//
// * For both component and low level use.
// * Most relevant functions ported.
// * Special functionality to get results in
//   two-dimensional array (type TQueryResult)
// * design time functionality




// So far the .dll / .so function library interface

///////////////////////////////////////
// THE ACTUAL TMyDB component :      //
///////////////////////////////////////

//TMyDB Component definition by ArTee:

const MY_DEFAULT_PORT=3306;

      //Resultset is a two-dimensional string array
      //implemented as :
      //TQueryResult   : Array of TQueryRow;
      //TQueryRow      : Array of String;
      //can be seen as : Array of Array of String
      //And used as    : ResultSet[x][y] (or: ResultSet[x,y])


type TQueryRow=Array of String;
     TQueryResult=Array of TQueryRow;

     TExtResult= record
                   IsNull:Boolean;
                   Data : ^String;
                   //Room for additional info,
                   //ToDo...
                 end;
     TExtQueryRow = Array of TExtResult;
     TExtQueryResult = Array of TExtQueryRow;

     //Some types redefined
     //Mainly PChar to String conversion:

{     TMySQLField = record
                     name,
                     table,
                     def: String;
                     _type: Byte;
                     length: Integer;
                     max_length: Integer;
                     flags: Integer;
                     decimals: Integer;
                   end;

     TMySQLFields = array of TFieldDesc;}

    TMyDB = class (TSQLDB)
    private
    function GetHasResult: Boolean;
    procedure SetEmbedded(const Value: Boolean);
    protected
      MyHandle:MySQL;
      PMyHandle:PMySQL;
      FLibrary:String;
      FHostInfo:String;
      FInfo:String;
      FRealConnect:Boolean;
      FUnixSock:String;
      FConnectOptions:Integer;
      FEmbedded: Boolean;
      procedure StoreResult(Res: PMYSQL_RES);
    public
      constructor Create (AOwner:TComponent); override;
      destructor Destroy; override;
      function Query (SQL:String):Boolean; override;
      function Connect (Host, User, Pass:String; DataBase:String=''):Boolean; override;
      procedure Close; override;
      function ExplainTable (Table:String): Boolean; override;
      function ShowCreateTable (Table:String): Boolean; override;
      function DumpTable (Table:String): Boolean; override;
      function DumpDatabase (Table:String): Boolean; override;

       //typical MySQL functions:
      function SelectDatabase(Database:String):Boolean;
//       function GetSelectedDatabase:String;
      procedure SetDatabase(Database:String);
      function CreateDatabase(Database:String):Boolean; virtual;
      function DropDatabase(Database:String):Boolean;
      procedure ListDatabases(wildcard:String='');
      procedure ListTables(wildcard:String='');
      procedure ListFields(table:String; wildcard:String='');
      procedure ListProcesses;
      function ShutDown:Boolean;
      function Kill (Pid:Integer):Boolean; //Kill specific process
      procedure SetPort (Port:Integer);
      procedure SetRealConnect(DoRealConnect:Boolean);
      function Ping:Boolean; //See if server is alive
      function GetLastError:String;
      function GetServerInfo:String;
      property DBHandle:MySQL read MyHandle; //Actual libmysql.dll / mysqlclient.so handle, use it if you want to call functions yourself
      property HasResult:Boolean read GetHasResult;// write FHasResult; //Queryhas valid result set
      property ServerInfo:String read GetServerInfo; //additional server info
      property Info:String read FInfo;
      property HostInfo:String read FHostInfo;
      property UnixSock:String read FUnixSock write FUnixSock;
    published
//      property SharedLibrary:String read FLibrary write FLibrary;
      property Embedded: Boolean read FEmbedded write SetEmbedded;
      property RealConnect:Boolean read FRealConnect write SetRealConnect;
    end;



implementation


//TMyDB has a constructor. Set some variabels to default, nothing more...
constructor TMyDB.Create;
begin
  FLibrary:=DEFAULT_DLL_LOCATION;
  {$IFDEF WIN32}
  DLL:=FLibrary;
  {$ENDIF}
  FHost:='localhost';
  FPort:=MY_DEFAULT_PORT;
  FActive:=False;
  FActivateOnLoad:=False;
  FRealConnect:=False;
  FConnectOptions:=_CLIENT_COMPRESS or _CLIENT_CONNECT_WITH_DB;
  FDataBaseType := dbMySQL;
  PrimaryKey := 'auto_increment primary key';
  inherited Create(AOwner);
end;

destructor TMyDB.Destroy;
begin
  if Active then
    Close;
  inherited Destroy;
end;

procedure TMyDB.Close;
begin
  if FActive then
    try
      mysql_close(@MyHandle);
    except
      raise Exception.Create('An error occured while closing');
    end;
  FActive:=False;
  if Assigned(FOnClose) then
    FOnClose(Self);
end;

function TMyDB.Connect(Host, User, Pass:String; DataBase:String):Boolean;
var AH:PMySQL;
begin

  {$IFDEF WIN32}
  //Allow user to change shared library
  if FLibrary<>'' then
    DLL_Client:=FLibrary;

  if (hDLLClient=0) and not MySQLLoadLib(FEmbedded) then
    exit
  else
    FDllLoaded := True;
  {$ENDIF}

  //Succesfully loaded
  if FActive then Close;  //Close if already active
  if assigned(mysql_init) then
    PMyHandle := mysql_init(@MyHandle)
  else
    exit;
  FDataBase := DataBase;

  //Enable realconnect by default, not overridable...
  FRealConnect := True;
//  if MyHandle=nil then
//    exit;
  if FRealConnect then
    try
      PMyHandle:= mysql_real_connect(@MyHandle, PChar(String(Host)), PChar(String(User)), PChar(String(Pass)),
                 PChar(String(FDataBase)), FPort, nil {PChar(String(FUnixSock))}, 0{ FConnectOptions});
      FActive := PMyHandle<>nil;
      if not FActive then
        begin
          FCurrentSet.FLastErrorText := MyHandle._net.last_error;
          if pos(#0, FCurrentSet.FLastErrorText)>0 then //probably is
            FCurrentSet.FLastErrorText := copy (FCurrentSet.FLastErrorText, 1, pos(#0, FCurrentSet.FLastErrorText)-1);
          FCurrentSet.FLastError := MyHandle._net.last_errno;
          LogError;
        end
      else
        FCurrentSet.FLastError := 0;
    except
      FActive:=False;
    end
  else
    begin
      AH{PMyHandle}:=mysql_connect(@MyHandle, PChar(Host), PChar(User), PChar(Pass));
      FActive := PMyHandle<>nil;
      //Select database if assigned:
      if FActive and (FDataBase<>'') then
        mysql_select_db(@MyHandle, PChar(FDataBase));
    end;

  Result := FActive;
  //Don't know why, but hangs if in componentstate:
  if FActive and not (csDesigning in ComponentState) then
    begin
      //Fill in some variables:
      //If compiled with wrong version, the MYSQL type may be wrong,
      //Therefore an try-except:
      try
      //It's too risky. i go figure out what's happening.

      {
      if Assigned (MyHandle.server_version) then
        FVersion:=MyHandle.server_version;
      if Assigned (MyHandle.Charset) then
        FEncoding:=MyHandle.charset;
      }

      {
      if Assigned (MyHandle.Host_Info) then
        FHostInfo:=MyHandle.host_info;
      }
      {
      if Assigned (MyHandle.Info) then
        FInfo:=MyHandle.info;
      }
      except
        //Wrong DLL. Adjust compiler switches of unit MyDB. :(
      end;
    end;

  if FActive and Assigned(FOnOpen) then
    FOnOpen(Self);
end;

//An active property was added to allow
//database-access in development state ;)

//Quite direct MySQL functions:
function TMyDB.CreateDatabase(Database:String):Boolean;
begin
  if FActive and assigned (mysql_create_db) then
    Result := (0=mysql_create_db(@MyHandle, PChar(Database)))
  else
    Result := False;
end;

function TMyDB.DropDatabase(Database:String):Boolean;
begin
  if FActive and assigned (mysql_drop_db) then
    Result := (0=mysql_drop_db(@MyHandle, PChar(Database)))
  else
    Result := False;
  if Result and
     (lowercase(FDataBase) = lowercase(DataBase)) then
    FDatabase:='';
end;

function TMyDB.SelectDatabase(Database:String):Boolean;
begin
  if FActive and assigned (mysql_select_db) then
    Result := (0 = mysql_select_db(@MyHandle, PChar(Database)))
  else
    Result := False;
  if Result then FDatabase:=Database;
end;

function TMyDB.Kill(Pid:Integer):Boolean;
begin
  if FActive and assigned (mysql_kill) then
    Result := (0=mysql_kill(@MyHandle, Pid))
  else
    Result := False;
end;

function TMyDB.Ping: Boolean;
begin
  Result:=False;
  if FActive and assigned (mysql_ping) then
    Result:=(mysql_ping(@MyHandle)<>0);
end;

function TMyDB.ShutDown: Boolean;
begin
  if FActive then
    Result := (0=mysql_shutdown(@MyHandle))
  else
    Result := False;
end;


//This is where the results from a query are stored in delphi string-arrays

(* Old version as of TMyDB 0.2, leave for comparance...
   //ToDo: Add universal field record...//
procedure TMyDB.StoreResult;
//Loop all rows from a result set and put fields in 2D-array
var i,j:Integer;
    row:mysql_row;
    fields:Pmysql_fields;
    TotMem:Integer;
begin
  FHasResult:=False;
  if Res<>nil then
    begin
      FHasResult:=True;
      //fetch rows
      SetLength (FQueryResult, Integer(res^.row_count));
      SetLength (FExtQueryResult, Integer(res^.row_count));
      TotMem:=0;
      for i:=0 to res^.row_count-1 do
        begin
          //Check ranges; break if rowlimit or memory limit reached:
          if ((FFetchRowLimit<>0) and ((i+1)>=FFetchRowLimit)) or
             ((FFetchMemoryLimit<>0) and (TotMem>=FFetchMemoryLimit)) then
            break;
          //Fetch a row:
          row:=mysql_fetch_row(res)^;
          SetLength (FQueryResult[i], res^.field_count);
          SetLength (FExtQueryResult[i], res^.field_count);
          for j:=0 to res^.field_count-1 do
            begin
              if row[j]=nil then
                FExtQueryResult[i,j].IsNull :=True;
              FQueryResult[i,j]:=row[j];
              FExtQueryResult[i,j].Data := @FQueryResult[i,j];
              TotMem:=16+length(FQueryResult[i,j]);
            end;
        end;

      //Leave first column, first row, as single string:
      if high(FQueryResult)>=0 then
        FResultAsString := FQueryResult[0][0]
      else
        FResultAsString := '';

      //Fetch field names
      FFieldNames.Clear;

      fields:=mysql_fetch_fields(res);
      SetLength (FFields, mysql_num_fields(res));
      for i:=0 to mysql_num_fields(res)-1 do
        begin
          FFieldNames.Add(fields^[i].Name);
          with FFields[i] do begin
            //Copy data mainly for PChar/String converting
            //Makes field info available after resource handle is closed!
            name:=fields[i].name;
            table:=fields[i].table;
            def:=fields[i].def;
            datatype:=fields[i].enum_field_type;
            length:=fields[i].length;
            max_length:=fields[i].max_length;
            flags:=fields[i].flags;
            decimals:=fields[i].decimals;
          end;
        end;
      //Some more vars:
      FNumFields:=mysql_num_fields(res);
      FRowsAffected:=MyHandle.affected_rows;
      FLastInsertID:=MyHandle.insert_id;
      mysql_free_result(res);
      FHasResult:=True;
      if Assigned (OnSuccess) then
        OnSuccess(Self);
    end
  else //Invalid Result
    begin
      SetLength (FQueryResult, 0);
      FResultAsString:='';
//      FResultAsText:='';
//      FResultAsHTML:='';
      FNumFields:=0;
      FRowsAffected:=-1;
      FLastInsertID:=-1;
      FFieldNames.Clear;
      FHasResult:=False;
      if Assigned (OnError) then
        OnError (Self);
    end;
end;
*)

procedure TMyDB.StoreResult;
//Loop all rows from a result set and put fields in 2D-array
var i, j, ri: Integer;
    row: mysql_row;
    fields: Pmysql_fields;
    FieldDesc: TFieldDesc;
    TotMem: Integer;
    R: TResultRow;
begin
  with FCurrentSet do
    begin
      FHasResult:=False;
      if Res<>nil then
        begin
          FHasResult:=True;
          //fetch rows
          TotMem:=0;
          FRowCount := res^.row_count;
          fields:=mysql_fetch_fields(res);

          for i:=0 to res^.row_count-1 do
            begin
              if FCallBackOnly then
                ri:=0 //only 1 row needed
              else
                begin
                  ri := i;
                  //Check ranges; break if rowlimit or memory limit reached:
                  if ((FFetchRowLimit<>0) and ((i+1)>=FFetchRowLimit)) or
                     ((FFetchMemoryLimit<>0) and (FQuerySize>=FFetchMemoryLimit)) then
                    break; //mem limit exceeded...
                end;

              //Fetch a row:
              row:=mysql_fetch_row(res)^;
              if ri<FRowList.Count then
                begin
                  R := TResultRow(FRowList[ri]);
                  R.Clear;
                  R.FNulls.Clear;
                end
              else
                begin
                  R := TResultRow.Create;
                  R.FFields := FFields; //copy pointer to ffields array
                  FRowList.Add(R);
                end;

    //          SetLength (FQueryResult[i], res^.field_count);
    //          SetLength (FExtQueryResult[i], res^.field_count);

//              for j:=0 to res^.field_count-1 do
             for j:=0 to mysql_num_fields(res) - 1 do
                begin
                  if Assigned (row[j]) then
                    R.Add(row[j])
                  else
                    R.Add('');
                  R.FNulls.Add(Pointer(Integer(row[j]<>nil)));
                  inc (FQuerySize, length(String(row[j])));
                end;
              if Assigned (FOnFetchRow) then
                try
                  FOnFetchRow (Self, R);
                except end;

            end;


          //Leave first column, first row, as single string:
    {      if high(FQueryResult)>=0 then
            FResultAsString := FQueryResult[0][0]
          else
            FResultAsString := '';}

          //Fetch field names
    //      FFieldNames.Clear;


    //      SetLength (FFields, mysql_num_fields(res));
          for i:=0 to mysql_num_fields(res)-1 do
            begin
              FieldDesc := TFieldDesc.Create;
              FFields.AddObject(fields^[i].Name, FieldDesc);
              with FieldDesc do begin
                //Copy data mainly for PChar/String converting
                //Makes field info available after resource handle is closed!
                name:=fields[i].name;
                table:=fields[i].table;
//                def:=fields[i].def;   //sorry...
                datatype:=fields[i].enum_field_type;
    //            length:=fields[i].length;
                max_length:=fields[i].max_length;
                flags:=fields[i].flags;
                decimals:=fields[i].decimals;
              end;
            end;
          //Some more vars:
          FColCount:=mysql_num_fields(res);
          FRowsAffected:=MyHandle.affected_rows;
          FLastInsertID:=MyHandle.insert_id;
          mysql_free_result(res);
          FHasResult:=True;
          if Assigned (OnSuccess) then
            OnSuccess(Self);
        end
      else //May be invalid result or just no result
        begin
          Clear;
          FLastErrorText := mysql_error(@MyHandle);
          FLastError := mysql_errno(@MyHandle);
          if (FLastError<>0) and (Assigned (OnError)) then
            OnError (Self);
        end;
      //
      //some differences in mysql4, these params are no longer available in the mysql handle (...?)
      FLastInsertID := mysql_insert_id (@MyHandle);
    //  FRowsAffected := mysql_rows_affected (@MyHandle);
  end;
end;


//This is the main function to perform a query:
function TMyDB.Query (SQL: String): Boolean;
begin
  Result := False;
  if not FActive then
    SetActive(True); //Try once if client just performs query

  Clear;

  with FCurrentSet do
    begin
      FHasResult := False;

      if not FActive then
        exit; //sorry... nothing to do here, handle is invalid.

      if SQL='' then //clear the results:
        begin
          StoreResult (nil);
          exit;
        end;

      if FActive then
        begin
          //Allow user to view or edit query:
          FSQL:=SQL;
          if Assigned (FOnBeforeQuery) then
            FOnBeforeQuery(Self, FSQL);
          SQL:=FSQL;

          //Perform actual query:
          if 0=mysql_query(@MyHandle, PChar(SQL)) then
            begin
              StoreResult(mysql_store_result(@MyHandle));
              Result := true;
              FLastError := 0;
              FLastErrorText := '';
              FHasResult := True;
            end
          else
            begin
              //StoreResult is able to handle errors and will call OnError as well
              //Calling it with nill forces a result cleanup:
              StoreResult(nil);
              FLastErrorText := mysql_error(@MyHandle); //MyHandle._net.last_error;
              if pos(#0, FLastErrorText)>0 then //probably is
                FLastErrorText := copy (FLastErrorText, 1, pos(#0, FLastErrorText)-1);
              FLastError := mysql_errno(@MyHandle); //MyHandle._net.last_errno;
//              if Assigned (FOnError) then
//                FOnError(Self);
              LogError;
            end;
        end;
  end;
end;


//Common libmysql / libmysqlclient functions:
procedure TMyDB.ListDatabases;
begin
  if FActive then
    StoreResult(mysql_list_dbs(@MyHandle, PChar(wildcard)));
end;

procedure TMyDB.ListTables;
begin
  if FActive then
    StoreResult(mysql_list_tables(@MyHandle, PChar(wildcard)));
end;

procedure TMyDB.ListProcesses;
begin
  if FActive then
    StoreResult(mysql_list_processes(@MyHandle));
end;

procedure TMyDB.ListFields;
begin
  if FActive then
    StoreResult(mysql_list_fields(@MyHandle, PChar(table), PChar(wildcard)));
end;

function TMyDB.GetServerInfo: String;
begin
  if FActive then
    Result:=mysql_get_server_info(@MyHandle)
  else
    Result:='Inactive';
end;

function TMyDB.GetLastError: String;
//var B:array[01..MYSQL_ERRMSG_SIZE] of Char;
begin
  Result := FCurrentSet.FLastErrorText;
  {
  if FActive then
    begin
      SetLength(Result, 200);
      Result:=myhandle._net.last_error;
    end;
  }
end;

//TMyDB control functions:
procedure TMyDB.SetPort;
begin
  if (Port<=0) or (Port>65535) then //Simply don't accept value
    exit;
  if Port<>MY_DEFAULT_PORT then //Force real connect:
    FRealConnect:=True;
  FPort:=Port;
end;

procedure TMyDB.SetRealConnect;
begin
  if not DoRealConnect then //Only connect to default port:
    FPort:=MY_DEFAULT_PORT;
  FRealConnect:=DoRealConnect;
end;

procedure TMyDB.SetDatabase;
begin
  if FActive then
    begin
      if SelectDataBase(DataBase) then
        FDataBase :=DataBase;
    end
  else
    FDataBase := Database;
end;


function TMyDB.DumpDatabase(Table: String): Boolean;
begin

end;

function TMyDB.DumpTable(Table: String): Boolean;
begin

end;

function TMyDB.ExplainTable(Table: String): Boolean;
begin
  Result := FormatQuery ('explain table %q', [Table]);
end;

function TMyDB.ShowCreateTable(Table: String): Boolean;
begin

end;

function TMyDB.GetHasResult: Boolean;
begin

end;

procedure TMyDB.SetEmbedded(const Value: Boolean);
begin
  FEmbedded := Value;
  if FEmbedded then
    FLibrary := MYSQLD_DLL_LOCATION
  else
    FLibrary := DEFAULT_DLL_LOCATION;
end;

end.
